/* \file pipe16.cpp
 *
 *  RiSC-16 Pipeline Simulator
 */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <string>
using namespace std;
#include "strtoken.h"

int mem[512];
int nwords = 0;

void read_memory(FILE *in);
void show_memory(FILE *out);
int mem_check(int addr);
void init_registers();
void show_registers(FILE *out);
void simulate(FILE *out);
void pipeline(FILE *out);

int main(int argc, char *argv[])
{
	FILE *in, *out;
	string filename;
	errno_t nerr;
	if (argc<2) {
		cout << "usage: pipe16 hexfile [output file]\n";
		return EXIT_FAILURE;
	}
	filename = argv[1];
	nerr = fopen_s(&in,filename.c_str(),"r");
	if (nerr) {
		cout << "file: " << filename << " not found\n";
		return EXIT_FAILURE;
	}
	if (argc>2) {
		filename = argv[2];
		nerr = fopen_s(&out,filename.c_str(),"wt");
		if (nerr) {
			cout << "file: " << filename << " could not be opened\n";
			return EXIT_FAILURE;
		}
	}
	else out = stdout;
	// single cycle simulation
	read_memory(in);
	fprintf(out,"<html>\n");
	simulate(out);
	fprintf(out,"<h3>Registers</h3>\n");
	show_registers(out);
	fprintf(out,"<h3>Memory</h3>\n");
	show_memory(out);
	// pipeline
	rewind(in);
	read_memory(in);
	pipeline(out);
	fclose(in);
	fprintf(out,"<h3>Registers</h3>\n");
	show_registers(out);
	fprintf(out,"<h3>Memory</h3>\n");
	show_memory(out);
	return EXIT_SUCCESS;
}

void show_memory(FILE *out)
{
	int i,flag=0;
	fprintf(out,"<p><table border>\n");
	fprintf(out,"<p><tr bgcolor=CCCCFF><th>addr");
	for (i=0; i<16; i++) fprintf(out,"<th>%X",i);
	fprintf(out,"\n");
	for (i=0; i<nwords; i++) {
		flag = i%16;
		if (flag==0) fprintf(out,"\n<tr align=center><td bgcolor=FFFFCC>%04X",i);
		fprintf(out,"<td>%04X",mem[i]);
		if (flag==7) fprintf(out,"\n");
	}
	while (flag++ < 15) fprintf(out,"<td><br>");
	fprintf(out,"\n</table>\n");
	
}

void read_memory(FILE *in)
{
	char buf[512];
	char *inp, *next;
	int field;
	bool done = false;
	init_registers();
	nwords = 0;
	while (!done) {
		fgets(buf,511,in);
		if (feof(in)) break;
		next = buf;
		while (*next==' ') next++;
		while (next) {
			inp = next;
			while (*inp==' ') inp++; // eat blanks
			next = strqtrm(inp,' ');
			sscanf_s(inp,"%x",&field);
			//if (field>0) printf("%04X\n",field);
			mem[nwords++] = field;
		}
	}
}



int mem_check(int addr)
{
	if (addr<0 || addr>=nwords) {
		fprintf(stderr,"address %x out of range\n",addr);
		return 0;
	}
	return 1;
}


int regs[] = {0xBAD0,0x1111,0x2222,0x3333,0x4444,0x5555,0x6666,0x7777};

int reg(int n)
{
	if (n>0) return regs[n];
	return 0;
}

void init_registers()
{
	regs[1] = 0x1111;
	regs[2] = 0x2222;
	regs[3] = 0x3333;
	regs[4] = 0x4444;
	regs[5] = 0x5555;
	regs[6] = 0x6666;
	regs[7] = 0x7777;
	regs[8] = 0x8888;
}


void show_registers(FILE *out)
{
	int i;
	fprintf(out,"<p><table border>\n");
	fprintf(out,"<p><tr bgcolor=CCCCFF><th>format");
	for (i=1; i<8; i++) fprintf(out,"<th>r%d",i);
	fprintf(out,"\n");
	fprintf(out,"<tr align=center><th>hex");
	for (i=1; i<8; i++) {
		fprintf(out,"<td>%04X",reg(i));
	}
	fprintf(out,"<tr align=center><th>decimal");
	for (i=1; i<8; i++) {
		fprintf(out,"<td>%d",(short)reg(i));
	}
	fprintf(out,"\n</table>\n");
}

string cmd[] = { "add", "addi", "nand", "lui", "sw", "lw", "bne", "jalr"};

const int ADD = 0, ADDI = 1, NAND = 2,  LUI = 3;
const int  SW = 4,   LW = 5,  BNE = 6, JALR = 7;

const int HALT = 0xe071;

class cycle {
public:
	int pc;
	int op;
	int rd;
	int qd;
	int addr;
	void set(int pci, int opi, int rdi, int qdi);
	void show(FILE *out, int ncycle);
};

void cycle::set(int pci, int opi, int rdi, int qdi)
{
	pc = pci;
	op = opi;
	rd = rdi;
	qd = qdi;
}

void cycle::show(FILE *out, int ncycle)
{
	fprintf(out,"<tr><td bgcolor=FFFFCC>%d<td>%02X",ncycle+1,pc);
	fprintf(out,"<td>%s<td>%d<td>%04X<td>",
		  cmd[op].c_str(),rd,qd);
	if (op==LW || op==SW) fprintf(out,"%04X\n",addr);
	else fprintf(out,"<br>\n");
}

const int MAX_CYCLES = 256;
cycle steps[MAX_CYCLES];
int nsteps = 0;
	

void simulate(FILE *out)
{
	int pc, nextpc;
	int inst, seim, im10, maddr;
	int opcode, rd=0, ra=0, rb=0;
	int qa, qb, qd=0;
	bool flag, done = false;
	if (out) {
		fprintf(out,"<h3>Simulation Results</h3>\n");
		fprintf(out,"<p><table border>\n");
		fprintf(out,"<tr bgcolor=CCCCFF><th>#<th>pc<th>op<th>rd<th>qd<th>addr\n");
	}
	nextpc = 0;
	while (!done) {
		inst = mem[nextpc];
		opcode = inst>>13;
		rd = (inst>>10)&0x7;
		ra = (inst>>7)&0x7;
		rb = inst&0x7;
		seim = inst&0x7F;
		im10 = inst&0x3ff;
		if (seim>>6) seim = (0xFFFFFF80)| seim;
		qa = reg(ra);
		qb = reg(rb);
		maddr = qa + seim;
		if (inst==HALT) break;
		pc = nextpc;
		nextpc++;
		switch (opcode) {
			case ADD:
				qd = qa + qb;
				break;
			case ADDI:
				qd = qa + seim;
				break;
			case NAND:
				qd = ~(qa & qb);
				qd &= 0xFFFF;
				break;
			case LUI:
				qd = (im10<<6);
				break;
			case SW: // SW
				qd = reg(rd);
				if (mem_check(maddr)) mem[maddr] = qd;
				break;
			case LW:
				if (mem_check(maddr)) qd = mem[maddr];
				break;
			case BNE:
				flag = reg(rd)!=qa;
				if (flag) {
					nextpc += seim;
					qd = nextpc;
				}
				else qd = 0;
				break;
			case JALR:
				qd = nextpc;
				nextpc = qa;
				break;
				
		}
		qd &= 0xFFFF;
		if (rd && opcode!=SW && opcode!=BNE) regs[rd] = qd;
		int ndx = nsteps;
		steps[ndx].set(pc,opcode,rd,qd);
		if (opcode==SW || opcode==LW) steps[ndx].addr = maddr;
		if (out) steps[ndx].show(out,ndx);
		if (++nsteps>=MAX_CYCLES) break;
	}
	if (out) fprintf(out,"</table>\n");
}

void showpc(FILE *out, int pc, bool color)
{
	if (color) fprintf(out,"<td bgcolor=E0E0E0");
	else fprintf(out,"<td");
	fprintf(out," >%04x",pc);
}

class IFID {
public:
	bool active;
	bool color;
	bool halt;
	int instr;
	int pc;
	int op;
	int rd, ra, rb;
	void show(FILE *out);
	void bubble();
};

void IFID::show(FILE *out)
{
	if (active) {
		showpc(out,pc,color);
		fprintf(out,"<br>%s",halt? "halt":cmd[op].c_str());
	}
	else fprintf(out,"<TD><BR>");
}

void IFID::bubble()
{
	active = false;
	halt = false;
	instr = 0;
	op = 0;
	rd = 0;
}

class IDEX {
public:
	bool active;
	bool color;
	bool halt;
	int pc;
	int instr;
	int rd, ra, rb;
	int qa, qb;
	void show(FILE *out);
	void bubble();
	void setreg();
};

void IDEX::bubble()
{
	active = false;
	halt = false;
	instr = 0;
	rd = 0;
}

void IDEX::setreg()
{
	qa = reg(ra);
	qb = reg(rb);
}

void IDEX::show(FILE *out)
{
	if (active) {
		showpc(out,pc,color);
		fprintf(out,"<br>%d %d %d\n",rd,ra,rb);
	}
	else fprintf(out,"<td><br>");
}

class EXWB {
public:
	bool active;
	bool color;
	int pc;
	int instr;
	bool ma;
	int maddr;
	int jump;
	int rd;
	int qd;
	void show(FILE *out);
	void showmem(FILE *out);
};

void EXWB::showmem(FILE *out)
{
	int op = instr>>13;
	if (op==LW) fprintf(out,"qd = %04x from mem[0x%x]\n",qd,maddr);
	if (op==SW) fprintf(out,"mem[0x%x] = %04x\n",maddr,qd);
}

void EXWB::show(FILE *out)
{
	int opcode = instr>>13;
	bool flag = (opcode==LW || opcode==SW);
	if (active) {
		showpc(out,pc,color);
		fprintf(out,"<br>%04x",(flag? maddr: qd));
	}
	else fprintf(out,"<td><br>");
}

class WBXX {
public:
	bool active;
	bool color;
	int pc;
	int instr;
	int rd;
	int qd;
	void show(FILE *out);
};

void WBXX::show(FILE *out)
{
	if (active) {
		showpc(out,pc,color);
		fprintf(out,"<br>%04x",qd);
	}
	else fprintf(out,"<td><br>");
}


IFID IFID_old, IFID_new;
IDEX IDEX_old, IDEX_new;
EXWB EXWB_old, EXWB_new;
WBXX WBXX_old, WBXX_new;

#define FORWARDING

void pipeline(FILE *out)
{
	int nextpc, ncycle=0;
	int ndx = 0;
	int inst, opcode;
	int seim, im10;
	int ra, rb, rd;
	int qa, qb, qd;
	int jump;
	bool done = false;
	if (out) {
		fprintf(out,"<h3>Pipeline Results</h3>\n");
		fprintf(out,"<p><table border>\n");
		fprintf(out,"<tr bgcolor=CCCCFF><th>#<th>IF<th>ID<th>EX<th>WB<th>ok\n");
	}
	nextpc = 0;
	ncycle = 0;
	while (!done) {
		// IF stage
		inst =  mem[nextpc];
		opcode = inst>>13;
		IFID_new.active = !IFID_old.halt;
		IFID_new.color = (ncycle&3)==2;
		IFID_new.instr = inst;
		IFID_new.halt = (inst==HALT) || IFID_old.halt;
		IFID_new.pc = nextpc;
		IFID_new.op = opcode;
		rd = (inst>>10)&0x7;
		ra = (inst>>7)&0x7;
		rb = inst&0x7;
		bool flag = opcode==SW || opcode==BNE;
		IFID_new.rd = flag? 0: rd;
		IFID_new.ra = ra;
		IFID_new.rb = (opcode==ADD || opcode==NAND)? rb:
			      (flag? rd: 0);
		nextpc++;
		// ID stage
		IDEX_new.active = IFID_old.active;
		IDEX_new.color = IFID_old.color;
		IDEX_new.instr = IFID_old.instr;
		IDEX_new.halt = IFID_old.halt;
		IDEX_new.pc = IFID_old.pc;
		IDEX_new.rd = IFID_old.rd;
		IDEX_new.ra = IFID_old.ra;
		IDEX_new.rb = IFID_old.rb;
		IDEX_new.setreg();
		//  EX Stage
		EXWB_new.active = IDEX_old.active;
		EXWB_new.color = IDEX_old.color;
		EXWB_new.pc = IDEX_old.pc;
		inst = IDEX_old.instr;
		EXWB_new.instr = inst;
		EXWB_new.rd = IDEX_old.rd;
		opcode = inst>>13;
		seim = inst&0x7F;
		im10 = inst&0x3ff;
		if (seim>>6) seim = (0xFFFFFF80)| seim;
		qa = reg(IDEX_old.ra);
		qb = reg(IDEX_old.rb);
		// data forwarding
#if defined(FORWARDING)
		if (IDEX_old.ra == EXWB_old.rd && IDEX_old.ra !=0) {
			qa = EXWB_old.qd;
			fprintf(stderr,"(cycle %d) qa forwarded: %04x\n",ncycle+1,qa);
			fprintf(out,"<tr align=center><td><br>");
			fprintf(out,"<td bgcolor=FFFF00 colspan=5>qa forwarded: %04x\n",qa);
		}
		if (IDEX_old.rb == EXWB_old.rd && IDEX_old.rb !=0) {
			qb = EXWB_old.qd;
			fprintf(stderr,"(cycle %d) qb forwarded: %04x\n",ncycle+1,qb);
			fprintf(out,"<tr align=center><td><br>");
			fprintf(out,"<td bgcolor=FFFF00 colspan=5>qb forwarded: %04x\n",qb);
		}
#endif
		//
		int maddr = qa + seim;
		EXWB_new.maddr = maddr;
		jump = 0;
		EXWB_new.ma = false;
		switch (opcode) {
			case ADD:
				qd = qa + qb;
				break;
			case ADDI:
				qd = qa + seim;
				break;
			case NAND:
				qd = ~(qa & qb);
				qd &= 0xFFFF;
				break;
			case LUI:
				qd = (im10<<6);
				break;
			case SW: // SW
				EXWB_new.ma = true;
				qd = qb;
				if (mem_check(maddr)) mem[maddr] = qd;
				break;
			case LW:
				EXWB_new.ma = true;
				if (mem_check(maddr)) qd = mem[maddr];
				break;
			case BNE:
				if (qa!=qb) {
					qd = IDEX_old.pc + 1 + seim;
					nextpc = qd; // take branch
					IFID_new.bubble();
					IDEX_new.bubble();
					//fprintf(out,"<tr align=center><td><br>");
					//fprintf(out,"<td bgcolor=8080FF colspan=5>branch %04x\n",jump);
				}
				else qd = 0;
				break;
			case JALR:
				if (IDEX_old.halt) break;
				qd = IDEX_old.pc+1;
				jump = qa;
				nextpc = jump; // take jump
				// insert bubbles
				IFID_new.bubble();
				IDEX_new.bubble();
				//fprintf(out,"<tr align=center><td><br>");
				//fprintf(out,"<td colspan=5> nextpc %04x jump %04x\n",qd,qa);
				break;

		}
		qd &= 0xFFFF;
		EXWB_new.qd = qd;
		// WB Stage
		WBXX_new.active = EXWB_old.active;
		WBXX_new.color = EXWB_old.color;
		WBXX_new.pc = EXWB_old.pc;
		inst = EXWB_old.instr;
		WBXX_new.instr = inst;
		opcode = inst>>13;
		rd = EXWB_old.rd;
		qd = EXWB_old.qd;
		WBXX_new.rd = rd;
		WBXX_new.qd = qd;
		if (rd>0) regs[rd] = qd;
		//
		if (++ncycle>=MAX_CYCLES) break;
		if (inst==HALT) break;
		// show state
		fprintf(out,"<tr align=center><td>%d",ncycle);
		IFID_new.show(out);
		IDEX_new.show(out);
		EXWB_new.show(out);
		WBXX_new.show(out);
		if (WBXX_new.active) {
			qd = steps[ndx].qd;
			int pc = steps[ndx].pc;
			bool flag = WBXX_new.qd==qd &&
				    WBXX_new.pc==pc;
			if (flag) fprintf(out,"<td>");
			else fprintf(out,"<td bgcolor=FF0000>");
			fprintf(out,"%04x<br>%04x\n",pc,qd);
			ndx++;
		}
		else fprintf(out,"<td><br>\n");
		if (EXWB_new.ma) {
			fprintf(out,"<tr align=center><td>%d",++ncycle);
			fprintf(out,"<td colspan=5 align=center>");
			EXWB_new.showmem(out);
		}
		// change state
		IFID_old = IFID_new;
		IDEX_old = IDEX_new;
		EXWB_old = EXWB_new;
		WBXX_old = WBXX_new;
	}
	if (out) fprintf(out,"</table>\n");
}
		


void parse(FILE *in, FILE *out)
{
	char buf[512];
	char *inp, *next;
	int opcode, ra, rb, rc, field;
	int count, code, n;
	string token;
	bool done = false;
	/*
	 * Go through input file, line by line
	 */
	n = 0;
	while (!done) {
		fgets(buf,511,in);
		if (feof(in)) break;
		//fprintf(out,"%s",buf);
		if (*buf=='#') continue;
		next = buf;
		/*
		 * parse current line
		 */
		count = 0;
		opcode = ra = rb = rc = 0;
		while (next) {
			/*
			 * search for '<', and copy prior text to
			 * output stream
 */
			inp = next;
			next = strqtrm(inp,',');
			sscanf_s(inp,"%d",&field);
			//fprintf(out,"count %d field %d\n",count,field);
			switch (count) {
			case 0:
				token = inp;
				//fprintf(out,"%s %d\n",token.c_str(),opcode);
				count = 1;
				break;
			case 1:
				ra = field;
				count = 2;
				break;
			case 2:
				rb = field;
				count = 3;
				break;
			case 3:
				rc = field;
				count = 4;
				break;
			default:
				break;
			}
		}
		code = ((opcode&0x7)<<13) | ((ra&0x7)<<10) | ((rb&0x7)<<7) | (rc&0x7F);;
		fprintf(out,"%04X\n",code);
		n++;
#if defined(STUFF)
			/*
			 * next points to character after ',' (if ','
			 * was found).
			 */
			/*
			if (next) {
				/*
				*  isolate (in html) text up to the character '>'
				*/
				html = next;
				next = strqtrm(html,'>');
				/*
				*  check for known codes
				*/
				string token = html;
				char *value = strqtrm(html,' ');
				if (token.substr(0,7).compare("include")==0) {
					include(out,value);
					continue; // go to next line
				}
				else if (token.substr(0,4).compare("copy")==0) {
					copy(out,value);
					continue; // go to next line
				}
				else if (token.substr(0,6).compare("insert")==0) {
					parse_insert(value,out);
				}
				/*
				 *  otherwise just echo the html code
				 */
				else fprintf(out,"<%s>",token.c_str());
			}
#endif
	}
	while (n++ < 32) fprintf(out,"0000\n");
}
